/*
 * Copyright (C) 2002-2003 Josh A. Beam
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define VERSION "0.1.2"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef __GLIBC__
#include <getopt.h>
#endif
#include "prtunnel.h"

unsigned int flags = 0;

extern unsigned char proxytype;
extern char *proxyhost;
extern unsigned short proxyport;

void show_usage_message(char *, FILE *);
void show_version_message();

extern void add_trusted_address(char *);
extern int fw_proxy(unsigned char[], unsigned short, char *, unsigned short, char *, char *);

int
main(int argc, char *argv[])
{
	int i;
	int ch;
	int localport;
	char *remotehost;
	int remoteport;
	char *username = NULL;
	char *password = NULL;
	static unsigned char localaddr[4] = { 0, 0, 0, 0 };

	for(i = 1; i < argc; i++) {
		if(strcmp(argv[i], "--help") == 0) {
			show_usage_message(argv[0], stdout);
			return 0;
		} else if(strcmp(argv[i], "--version") == 0) {
			show_version_message();
			return 0;
		}
	}

	while((ch = getopt(argc, argv, "hvDVcqt:H:P:T:u:p:")) != -1) {
		switch(ch) {
			case 'h':
				show_usage_message(argv[0], stdout);
				return 0;
			case 'v':
				show_version_message();
				return 0;
			case 'D':
				flags |= PRT_DAEMON;
				break;
			case 'V':
				flags |= PRT_VERBOSE;
				break;
			case 'c':
				flags |= PRT_COLOR;
				break;
			case 't':
				if(strcmp(optarg, "direct") == 0) {
					proxytype = PRT_DIRECT;
				} else if(strcmp(optarg, "http") == 0) {
					proxytype = PRT_HTTP;
				} else if(strcmp(optarg, "socks5") == 0) {
					proxytype = PRT_SOCKS5;
					proxyport = 1080;
				} else {
					fprintf(stderr, "Invalid proxy type `%s'\nRun `%s --help' for usage information.\n", optarg, argv[0]);
					return 1;
				}
				break;
			case 'H':
				proxyhost = optarg;
				break;
			case 'P':
				proxyport = atoi(optarg);
				break;
			case 'T':
				add_trusted_address(optarg);
				break;
			case 'u':
				username = optarg;
				break;
			case 'p':
				password = optarg;
				break;
		}
	}

	if((argc - optind) != 3) {
		if((argc - optind) == 0) {
			show_version_message();
			putchar('\n');
			show_usage_message(argv[0], stdout);
		} else {
			show_usage_message(argv[0], stderr);
		}
		return 1;
	}

	if(!proxyhost && proxytype != PRT_DIRECT) {
		fprintf(stderr, "No proxy hostname has been specified. You can specify one with the -H option.\nRun `%s --help' for more information.\n", argv[0]);
		return 1;
	}

	localport = atoi(argv[optind]);
	remotehost = argv[optind + 1];
	remoteport = atoi(argv[optind + 2]);

	if(fw_proxy(localaddr, localport, remotehost, remoteport, username, password) == -1) {
		fprintf(stderr, "prtunnel: Exiting because of error\n");
		return 1;
	}

	return 0;
}

void
show_usage_message(char *name, FILE *fp)
{
	fprintf(fp, "usage: %s [options] <local port> <remote host> <remote port>\n", name);
	fprintf(fp, "options:\n");
	fprintf(fp, "  -D\t\t\tRun as a daemon. prtunnel will run in the background\n\t\t\tand accept multiple TCP connections with this option.\n");
	fprintf(fp, "  -V\t\t\tVerbose output\n");
	fprintf(fp, "  -c\t\t\tUse color to differentiate between incoming\n\t\t\tand outgoing data in verbose output\n");
	fprintf(fp, "  -t <proxy type>\tSet proxy type. Valid types are http (default),\n\t\t\tsocks5, direct\n");
	fprintf(fp, "  -H <proxy host>\tSet proxy server hostname\n");
	fprintf(fp, "  -P <proxy port>\tSet proxy server port; defaults are 8080 for http,\n\t\t\t1080 for socks5\n");
	fprintf(fp, "  -T <address>\t\tAdd a trusted address. For security reasons, only\n\t\t\t127.0.0.1 is trusted by default. See the prtunnel\n\t\t\tman page or README file for more information.\n");
	fprintf(fp, "  -u <username>\t\tSet authentication username\n");
	fprintf(fp, "  -p <password>\t\tSet authentication password\n");
	fprintf(fp, "  -h, --help\t\tShow this help message\n");
	fprintf(fp, "  -v, --version\t\tShow version information\n");
}

void
show_version_message()
{
	printf("prtunnel "VERSION"\n");
	printf("The latest version can be found at http://joshbeam.com/software.html\n");
	printf("Copyright (C) 2002-2003 Josh A. Beam <josh@joshbeam.com>\n");
	printf("There is NO WARRANTY for this software.\n");
}
