/*
 * Copyright (C) 2002-2003 Josh A. Beam
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include "prtunnel.h"

extern int flags;

unsigned char proxytype = PRT_HTTP;
char *proxyhost = NULL;
unsigned short proxyport = 8080;

struct trusted_address *trust_addrs = NULL;
unsigned int num_trust_addrs = 0;

unsigned int bytes = 0; /* bytes transferred; rx for parent, tx for child */

/* make sure we see the external protocol-specific functions */
extern int direct_connect_to(char *, unsigned short, char *, char *);
extern int direct_read(int, char *, int);
extern int direct_send(int, char *, int);

extern int http_connect_to(char *, unsigned short, char *, char *);
extern int http_read(int, char *, int);
extern int http_send(int, char *, int);

extern int socks5_connect_to(char *, unsigned short, char *, char *);
extern int socks5_read(int, char *, int);
extern int socks5_send(int, char *, int);

/* stuff needed for binding a socket */
struct boundsocket {
	int fd;
	struct sockaddr_in sin;
	unsigned int len;
};

/*
 * this structure just holds file descriptors and
 * pointers to read/send functions, basically.
 */
struct fwconnection {
	int localfd;
	int remotefd;
	struct boundsocket bs;

	int (*connect_to)(char *, unsigned short, char *, char *);
	/*
	 * Currently the separate pointers for local and remote operations
	 * are not needed, but I'm keeping them here just in case. They would
	 * be needed for socks5 udp, for example.
	 */
	int (*local_read)(int, char *, int);
	int (*local_send)(int, char *, int);
	int (*remote_read)(int, char *, int);
	int (*remote_send)(int, char *, int);
};

static struct fwconnection *fwcp = NULL;
int isparent = -1;

static void set_sigchld_handler();

static void
wait_for_children(int signal)
{
	if(signal != SIGCHLD && signal != SIGTERM)
		return;

	waitpid(0, NULL, WNOHANG);
	set_sigchld_handler();
}

static void
set_sigchld_handler()
{
	signal(SIGCHLD, wait_for_children);
}

static void
proxy_kill(int signal)
{
	if(!fwcp) {
		fprintf(stderr, "Error: Trying to quit without a proper fwcp pointer. Oh well, quitting anyway, with exit(1).\n");
		exit(1);
	}

	if(signal == SIGTERM) {
		close(fwcp->localfd);
		close(fwcp->remotefd);
		fwcp->localfd = fwcp->remotefd = -1;

		if(isparent) {
			wait(NULL);
			fprintf(stderr, "%u bytes recieved\n", bytes);
		} else {
			unsigned char *addr = (unsigned char *)&(fwcp->bs.sin.sin_addr);

			fprintf(stderr, "Connection with %u.%u.%u.%u:%u closed\n%u bytes sent\n", addr[0], addr[1], addr[2], addr[3], ntohs(fwcp->bs.sin.sin_port), bytes);
		}
	}
}

/*
 * if the PRT_VERBOSE bit is set, print s; if the PRT_COLOR bit
 * is set, print s in color for outgoing data
 */
static void
print_data(char *s, int len, unsigned char outgoing)
{
	int i;

	if(!(flags & PRT_VERBOSE))
		return;

	if(flags & PRT_COLOR) {
		if(outgoing)
			printf("\033[1;31m");
		else
			printf("\033[0;0m");

		for(i = 0; i < len; i++)
			putchar(s[i]);

		if(outgoing)
			printf("\033[0;0m");
	} else {
		for(i = 0; i < len; i++) {
			if(i == 0) {
				if(outgoing)
					printf(">>> ");
				else
					printf("<<< ");
			}
			putchar(s[i]);
			if(s[i] == '\n' && (len - 1) > i) {
				if(outgoing)
					printf(">>> ");
				else
					printf("<<< ");
			}
		}
	}

	fflush(stdout);
}

/* bind a tcp socket to address:port */
static struct boundsocket
bind_to(unsigned char address[4], unsigned short port)
{
	struct boundsocket bs;

	bs.fd = socket(AF_INET, SOCK_STREAM, 0);
	if(bs.fd == -1)
		return bs;

	bs.sin.sin_port = htons(port);
	memcpy(&bs.sin.sin_addr, address, 4);
	bs.len = sizeof(bs.sin);

	if((bind(bs.fd, (struct sockaddr *)&bs.sin, bs.len)) == -1) {
		close(bs.fd);
		bs.fd = -1;
		return bs;
	}

	return bs;
}

/* this does exactly as the name implies */
void
add_trusted_address(char *s)
{
	int i;
	struct hostent *host;
	unsigned char bitcheck = 0;

	if(!s)
		return;

	for(i = 0; s[i] != '/' && s[i] != '\0'; i++);
	if(s[i] == '/') {
		int j = 0;
		char tmp[3];

		s[i] = '\0';
		i++;
		while(s[i] != '\0' && j < 2)
			tmp[j++] = s[i++];

		bitcheck = atoi(tmp);
		if(bitcheck >= 32) {
			fprintf(stderr, "Error: Bad bitcheck number (%d) for address %s; must be 0 to 31\n", bitcheck, s);
			return;
		}
	}

	host = gethostbyname(s);
	if(!host) {
		fprintf(stderr, "Error: Unable to resolve hostname %s for trusted addresses\n", s);
		return;
	}

	if(!trust_addrs) {
		trust_addrs = malloc(sizeof(struct trusted_address));
		if(!trust_addrs) {
			fprintf(stderr, "Error: Unable to allocate memory for trusted address\n");
			return;
		}
		num_trust_addrs = 0;
	} else {
		struct trusted_address *tmp;

		tmp = realloc(trust_addrs, sizeof(struct trusted_address) * (num_trust_addrs + 1));
		if(!tmp) {
			fprintf(stderr, "Error: Unable to allocate memory for trusted address\n");
			return;
		}
		trust_addrs = tmp;
	}

	i = num_trust_addrs;
	trust_addrs[i].address[0] = host->h_addr[0];
	trust_addrs[i].address[1] = host->h_addr[1];
	trust_addrs[i].address[2] = host->h_addr[2];
	trust_addrs[i].address[3] = host->h_addr[3];
	trust_addrs[i].bitcheck = bitcheck;

	num_trust_addrs++;

	fprintf(stderr, "Added trusted address %u.%u.%u.%u", host->h_addr[0], host->h_addr[1], host->h_addr[2], host->h_addr[3]);
	if(bitcheck)
		fprintf(stderr, ", comparing only the first %u bits", bitcheck);
	fprintf(stderr, "\n");
}

/* return 1 if the specified address is trusted; otherwise return 0 */
static int
is_trusted_address(struct sockaddr_in *sin, socklen_t socklen)
{
	int i;
	unsigned char *addr = (unsigned char *)&(sin->sin_addr);
	unsigned long bits, bits2;

	if(addr[0] == 127 && addr[1] == 0 && addr[2] == 0 && addr[3] == 1)
		return 1; /* we can trust ourselves, right? right? */

	for(i = 0; i < num_trust_addrs; i++) {
		if(addr[0] == trust_addrs[i].address[0] &&
		   addr[1] == trust_addrs[i].address[1] &&
		   addr[2] == trust_addrs[i].address[2] &&
		   addr[3] == trust_addrs[i].address[3])
			return 1;

		if(trust_addrs[i].bitcheck) {
			if(sizeof(bits) < 4) {
				fprintf(stderr, "Error: Bad long size; unable to compare bits for trusted address check\n");
				return 0;
			}

			bits = addr[0];
			bits <<= 8;
			bits |= addr[1];
			bits <<= 8;
			bits |= addr[2];
			bits <<= 8;
			bits |= addr[3];

			bits2 = trust_addrs[i].address[0];
			bits2 <<= 8;
			bits2 |= trust_addrs[i].address[1];
			bits2 <<= 8;
			bits2 |= trust_addrs[i].address[2];
			bits2 <<= 8;
			bits2 |= trust_addrs[i].address[3];

			bits >>= (32 - trust_addrs[i].bitcheck);
			bits2 >>= (32 - trust_addrs[i].bitcheck);

			if(bits == bits2)
				return 1;
		}
	}

	return 0;
}

/* setup function pointers in fwc for protocol type; return 0 if successful */
static int
setup_functions(struct fwconnection *fwc, unsigned char type)
{
	if(type == PRT_DIRECT) {
		fwc->connect_to = direct_connect_to;
		fwc->local_read = direct_read;
		fwc->local_send = direct_send;
		fwc->remote_read = direct_read;
		fwc->remote_send = direct_send;
	} else if(type == PRT_HTTP) {
		fwc->connect_to = http_connect_to;
		fwc->local_read = http_read;
		fwc->local_send = http_send;
		fwc->remote_read = http_read;
		fwc->remote_send = http_send;
	} else if(type == PRT_SOCKS5) {
		fwc->connect_to = socks5_connect_to;
		fwc->local_read = socks5_read;
		fwc->local_send = socks5_send;
		fwc->remote_read = socks5_read;
		fwc->remote_send = socks5_send;
	} else {
		return -1;
	}

	return 0;
}

/*
 * here's where the real work is done. after establishing
 * a connection, we fork a child process; the child process
 * handles forwarding of data from the client to the server,
 * parent process forwards data from the server to the client.
 *
 * oh, and in daemon mode, a child is forked for handling
 * each connection.
 */
int
fw_proxy(unsigned char localaddr[4], unsigned short localport,
         char *remotehost, unsigned short remoteport,
         char *username, char *password)
{
	struct fwconnection fwc;
	unsigned char *addr;
	int pid, otherpid = -1;
	char buf[512];
	int len;

	if(flags & PRT_DAEMON) { /* we're a daemon, so fork and return */
		int daemonpid;

		daemonpid = fork();
		if(daemonpid == -1) {
			fprintf(stderr, "Error: Couldn't fork daemon process\n");
			return -1;
		} else if(daemonpid) {
			fprintf(stderr, "prtunnel daemon started\n");
			return 0;
		}

		setsid();
		chdir("/");
	}

	if(setup_functions(&fwc, proxytype) != 0) {
		fprintf(stderr, "Error: Invalid protocol type\n");
		return -1;
	}

	if(num_trust_addrs == 0) { /* bind to 127.0.0.1, as only it's trusted */
		localaddr[0] = 127; localaddr[1] = localaddr[2] = 0; localaddr[3] = 1;
	}
	fwc.bs = bind_to(localaddr, localport);
	if(fwc.bs.fd == -1) {
		fprintf(stderr, "Error: Unable to bind socket. The port you specified (%u) may be reserved or already in use.\n", localport);
		return -1;
	}

	if(listen(fwc.bs.fd, 0) == -1) {
		fprintf(stderr, "Error: Unable to listen to socket\n");
		close(fwc.bs.fd);
		return -1;
	}
	fprintf(stderr, "Waiting for connection to port %u...\n", localport);

	/*
	 * if we're in daemon mode, we fork child processes to
	 * handle new connections so that the parent process can
	 * loop and keep accepting new connections
	 */
	for(;;) {
		set_sigchld_handler();

		for(;;) {
			fwc.localfd = accept(fwc.bs.fd, (struct sockaddr *)&fwc.bs.sin, &fwc.bs.len);
			if(fwc.localfd == -1 && errno == EINTR) /* child probably quit */
				continue;
			else
				break;
		}
		if(fwc.localfd == -1) {
			fprintf(stderr, "Error: Unable to accept connection (errno = %d)\n", errno);
			close(fwc.bs.fd);
			return -1;
		}

		if(!is_trusted_address(&fwc.bs.sin, fwc.bs.len)) {
			addr = (unsigned char *)&(fwc.bs.sin.sin_addr);
			fprintf(stderr, "Connection attempt from non-trusted %u.%u.%u.%u:%u. Closing it.\n", addr[0], addr[1], addr[2], addr[3], ntohs(fwc.bs.sin.sin_port));
			close(fwc.localfd);
		} else {
			if(flags & PRT_DAEMON) {
				int tmp_pid;

				tmp_pid = fork();
				if(tmp_pid == -1) {
					fprintf(stderr, "Error: Couldn't fork process for accepted connection\n");
					close(fwc.localfd);
				} else if(tmp_pid == 0) {
					break;
				}
			} else {
				break;
			}
		}
	}

	addr = (unsigned char *)&(fwc.bs.sin.sin_addr);
	fprintf(stderr, "Connection from %u.%u.%u.%u:%u accepted\n", addr[0], addr[1], addr[2], addr[3], ntohs(fwc.bs.sin.sin_port));

	fwc.remotefd = fwc.connect_to(remotehost, remoteport, username, password);
	if(fwc.remotefd == -1) {
		fprintf(stderr, "Error: Unable to connect to %s:%u via proxy\n", remotehost, remoteport);
		close(fwc.localfd);
		close(fwc.bs.fd);
		return -1;
	}
	fprintf(stderr, "Connected to %s:%u via proxy\n", remotehost, remoteport);
	fwcp = &fwc;

	pid = fork();
	if(pid == -1) {
		fprintf(stderr, "Error: fork failed\n");
		close(fwc.remotefd);
		close(fwc.localfd);
		close(fwc.bs.fd);
		return -1;
	}

	signal(SIGTERM, proxy_kill);

	if(pid == 0) { /* child */
		otherpid = getppid();
		isparent = 0;
		fprintf(stderr, "Now proxying data...\n");

		while((len = fwc.local_read(fwc.localfd, buf, 512)) > 0) {
			fwc.remote_send(fwc.remotefd, buf, len);
			bytes += len;
			print_data(buf, len, 1);
		}

		if(fwc.localfd != -1 && fwc.remotefd != -1) {
			/*
			 * we usleep a bit here because the other process
			 * might still have some data to be read. without
			 * this, the "Connection closed" message is often
			 * printed before the last bit of verbose data.
			 */
			usleep(500);
			kill(otherpid, SIGTERM);
			proxy_kill(SIGTERM);
		}
	} else { /* parent */
		otherpid = pid;
		isparent = 1;
		while((len = fwc.remote_read(fwc.remotefd, buf, 512)) > 0) {
			fwc.local_send(fwc.localfd, buf, len);
			bytes += len;
			print_data(buf, len, 0);
		}

		if(fwc.localfd != -1 && fwc.remotefd != -1) {
			usleep(500);
			kill(otherpid, SIGTERM);
			proxy_kill(SIGTERM);
		}
	}

	return 0;
}
